# PITCH VIZ PARAMS
FIELD_WIDTH = 1000
FIELD_HEIGHT = 700
FIELD_DIM = (106.0, 68.0)

FIELD_COLOR = "mediumseagreen"
FIELD_MARKINGS_COLOR = "White"

# FIELD_COLOR = 'White'
# FIELD_MARKINGS_COLOR = 'black'


PLAYERMARKERSIZE = 20
player_marker_args = {
    "Home": dict(
        mode="markers+text",
        marker_size=PLAYERMARKERSIZE,
        marker_line_color="white",
        marker_color="red",
        marker_line_width=2,
        textfont=dict(size=11, color="white"),
    ),
    "Away": dict(
        mode="markers+text",
        marker_size=PLAYERMARKERSIZE,
        marker_line_color="white",
        marker_color="#0570b0",
        marker_line_width=2,
        textfont=dict(size=11, color="white"),
    ),
}

event_player_marker_args = {
    "Home": dict(
        mode="lines+markers+text",
        marker_size=PLAYERMARKERSIZE,
        marker_line_color="white",
        marker_color="red",
        marker_line_width=2,
        line_color="red",
        textfont=dict(size=11, color="white"),
    ),
    "Away": dict(
        mode="lines+markers+text",
        marker_size=PLAYERMARKERSIZE,
        marker_line_color="white",
        marker_color="#0570b0",
        marker_line_width=2,
        line_color="#0570b0",
        textfont=dict(size=11, color="white"),
    ),
}



from collections import namedtuple

import plotly.graph_objects as go

from src.pitchly.params import FIELD_COLOR  # event_player_marker_args,; player_marker_args,
from src.pitchly.params import FIELD_DIM
from src.pitchly.params import FIELD_HEIGHT
from src.pitchly.params import FIELD_MARKINGS_COLOR
from src.pitchly.params import FIELD_WIDTH


class Pitch:
    def __init__(self):
        # ALL DIMENSIONS IN m
        self.border_dimen = (3, 3)  # include a border arround of the field of width 3m
        self.meters_per_yard = 0.9144  # unit conversion from yards to meters
        self.half_pitch_length = FIELD_DIM[0] / 2.0  # length of half pitch
        self.half_pitch_width = FIELD_DIM[1] / 2.0  # width of half pitch
        self.signs = [-1, 1]
        # Soccer field dimensions typically defined in yards, so we need to convert to meters
        self.goal_line_width = 8 * self.meters_per_yard
        self.box_width = 20 * self.meters_per_yard
        self.box_length = 6 * self.meters_per_yard
        self.area_width = 44 * self.meters_per_yard
        self.area_length = 18 * self.meters_per_yard
        self.penalty_spot = 12 * self.meters_per_yard
        self.corner_radius = 1 * self.meters_per_yard
        self.D_length = 8 * self.meters_per_yard
        self.D_radius = 10 * self.meters_per_yard
        self.D_pos = 12 * self.meters_per_yard
        self.centre_circle_radius = 10 * self.meters_per_yard

        self.point = namedtuple("point", ["x", "y"])
        self.centre = self.point(0.0, 0.0)

    def get_layout(self, time=None, frameID=None, frame_range=None, title=None):
        shapes = []

        mid_circle = dict(
            type="circle",
            x0=self.centre.x - self.centre_circle_radius,
            y0=self.centre.y - self.centre_circle_radius,
            x1=self.centre.x + self.centre_circle_radius,
            y1=self.centre.y + self.centre_circle_radius,
            line_color=FIELD_MARKINGS_COLOR,
            layer="below",
        )
        mid_line = dict(
            type="line",
            x0=self.centre.x,
            y0=self.centre.y - self.half_pitch_width,
            x1=self.centre.x,
            y1=self.centre.y + self.half_pitch_width,
            line_color=FIELD_MARKINGS_COLOR,
            layer="below",
        )
        mid_point = dict(
            type="circle",
            x0=self.centre.x - 0.4,
            y0=self.centre.y - 0.4,
            x1=self.centre.x + 0.4,
            y1=self.centre.y + 0.4,
            line_color=FIELD_MARKINGS_COLOR,
            fillcolor=FIELD_MARKINGS_COLOR,
            layer="below",
        )
        shapes.extend([mid_circle, mid_line, mid_point])

        for s in self.signs:

            # plot pitch boundary
            boundary1 = dict(
                type="line",
                x0=-self.half_pitch_length,
                y0=s * self.half_pitch_width,
                x1=self.half_pitch_length,
                y1=s * self.half_pitch_width,
                line_color=FIELD_MARKINGS_COLOR,
                layer="below",
            )

            boundary2 = dict(
                type="line",
                x0=s * self.half_pitch_length,
                y0=-self.half_pitch_width,
                x1=s * self.half_pitch_length,
                y1=self.half_pitch_width,
                line_color=FIELD_MARKINGS_COLOR,
                layer="below",
            )

            circle = dict(
                type="circle",
                x0=s * (self.centre.x + self.half_pitch_length - self.penalty_spot) - self.centre_circle_radius,
                y0=self.centre.y - self.centre_circle_radius,
                x1=s * (self.centre.x + self.half_pitch_length - self.penalty_spot) + self.centre_circle_radius,
                y1=self.centre.y + self.centre_circle_radius,
                line_color=FIELD_MARKINGS_COLOR,
                layer="below",
            )

            patch = dict(
                type="rect",
                x0=s * self.half_pitch_length,
                y0=-self.area_width / 2.0 - 1,
                x1=s * (self.half_pitch_length - self.area_length),
                y1=self.area_width / 2.0 + 1,
                line=dict(color=FIELD_COLOR, width=0),
                fillcolor=FIELD_COLOR,
                layer="below",
            )

            box = dict(
                type="rect",
                x0=s * self.half_pitch_length,
                y0=-self.area_width / 2.0,
                x1=s * (self.half_pitch_length - self.area_length),
                y1=self.area_width / 2.0,
                line=dict(color=FIELD_MARKINGS_COLOR, width=2),
                layer="below",
            )
            D = dict(
                type="rect",
                x0=s * self.half_pitch_length,
                y0=-self.box_width / 2.0,
                x1=s * (self.half_pitch_length - self.box_length),
                y1=self.box_width / 2.0,
                line=dict(color=FIELD_MARKINGS_COLOR, width=2),
                layer="below",
            )
            pen = dict(
                type="circle",
                x0=s * (self.half_pitch_length - self.penalty_spot) - 0.4,
                y0=-0.4,
                x1=s * (self.half_pitch_length - self.penalty_spot) + 0.4,
                y1=0.4,
                line_color=FIELD_MARKINGS_COLOR,
                fillcolor=FIELD_MARKINGS_COLOR,
                layer="below",
            )

            top_post = dict(
                type="rect",
                x0=s * self.half_pitch_length,
                y0=self.goal_line_width / 2.0 - 0.5,
                x1=s * (self.half_pitch_length - 0.5),
                y1=self.goal_line_width / 2.0 + 0.5,
                line=dict(color=FIELD_MARKINGS_COLOR, width=0),
                fillcolor=FIELD_MARKINGS_COLOR,
                layer="below",
            )

            bottom_post = dict(
                type="rect",
                x0=s * self.half_pitch_length,
                y0=-self.goal_line_width / 2.0 - 0.5,
                x1=s * (self.half_pitch_length - 0.5),
                y1=-self.goal_line_width / 2.0 + 0.5,
                line=dict(color=FIELD_MARKINGS_COLOR, width=0),
                fillcolor=FIELD_MARKINGS_COLOR,
                layer="below",
            )

            shapes.extend([boundary1, boundary2, circle, patch, box, D, pen, top_post, bottom_post])

            # set axis limits
        xmax = FIELD_DIM[0] / 2.0 + self.border_dimen[0]
        ymax = FIELD_DIM[1] / 2.0 + self.border_dimen[1]

        layout = go.Layout(
            title={
                "text": f"Time: [{time}] | FrameID: {frameID}",
                "y": 0.99,
                "x": 0.5,
                "xanchor": "center",
                "yanchor": "top",
                "font_color": "white",
                "font_size": 20,
            },
            hovermode="closest",
            autosize=True,
            width=FIELD_WIDTH,
            height=FIELD_HEIGHT,
            plot_bgcolor=FIELD_COLOR,
            xaxis=go.layout.XAxis(
                range=[-xmax, xmax], showgrid=False, zeroline=False, showticklabels=False, visible=False
            ),
            yaxis=go.layout.YAxis(
                range=[-ymax, ymax],
                showgrid=False,
                zeroline=False,
                showticklabels=False,
                visible=False,
                scaleanchor="x",
                scaleratio=1,
            ),
            margin=go.layout.Margin(l=0, r=0, b=0, t=0, pad=0),
            legend=dict(x=0.1, y=-0.01, orientation="h"),
        )

        layout["shapes"] = shapes

        if frame_range:
            layout["sliders"], layout["updatemenus"] = self.add_controls(frame_range)
            layout["title"] = None
            layout["margin"]["b"] = 100

        if title:
            layout["title_text"] = title

        return layout

    def plot_freeze_frame(self, data, title):
        fig_dict = {"data": [], "layout": {}, "frames": []}

        fig_dict["layout"] = self.get_layout(title=title)
        fig_dict["data"] = data

        fig = go.Figure(fig_dict)
        fig.show()

    def plot_frames_sequence(self, data, frames, frame_range, title):
        fig_dict = {"data": [], "layout": {}, "frames": []}

        fig_dict["layout"] = self.get_layout(frame_range=frame_range, title=title)
        fig_dict["data"] = data
        fig_dict["frames"] = frames

        fig = go.Figure(fig_dict)
        fig.show()

    def add_controls(self, frames_to_track):
        sliders_dict = {
            "active": 0,
            #     "yanchor": "top",
            #     "xanchor": "left",
            "currentvalue": {"font": {"size": 20}, "prefix": "Frame:", "visible": True, "xanchor": "right"},
            "transition": {"duration": 300, "easing": "cubic-in-out"},
            "pad": {"b": 0, "t": 0},
            #     "len": 0.9,
            #     "x": 0.1,
            #     "y": 0,
            "steps": [],
        }

        for frameID in frames_to_track:
            slider_step = {
                "args": [
                    [frameID],
                    {
                        "frame": {"duration": 300, "redraw": False},
                        "mode": "immediate",
                        "transition": {"duration": 0, "easing": "linear"},
                    },
                ],
                "label": frameID,
                "method": "animate",
            }
            sliders_dict["steps"].append(slider_step)

        updatemenus = {
            "buttons": [
                {
                    "args": [
                        None,
                        {
                            "frame": {"duration": 0, "redraw": False},
                            "fromcurrent": True,
                            "mode": "immediate",
                            "transition": {"duration": 0, "easing": "linear"},
                        },
                    ],
                    "label": ">",
                    "method": "animate",
                },
                {
                    "args": [
                        [None],
                        {"frame": {"duration": 0, "redraw": False}, "mode": "immediate", "transition": {"duration": 0}},
                    ],
                    "label": "||",
                    "method": "animate",
                },
            ],
            "direction": "left",
            # "font":{"size":10},
            # "pad": {"r": 40, "t": 50},
            "showactive": False,
            "type": "buttons",
            "xanchor": "right",
            "yanchor": "top",
            "x": 0.1,
            "y": 0,
        }

        return [sliders_dict], [updatemenus]

    def plot_event(self, data, title):
        fig_dict = {"data": [], "layout": {}, "frames": []}

        fig_dict["layout"] = self.get_layout(title=title)
        fig_dict["data"] = data

        fig = go.Figure(fig_dict)
        fig.show()



import numpy as np
import plotly.figure_factory as ff
import plotly.graph_objects as go
from scipy import signal
from tqdm.auto import tqdm

# from src.pitchly.params import FIELD_HEIGHT
# from src.pitchly.params import FIELD_MARKINGS_COLOR
# from src.pitchly.params import FIELD_WIDTH
# from src.pitchly.params import event_player_marker_args
# from src.pitchly.params import FIELD_COLOR
from src.pitchly.params import FIELD_DIM
from src.pitchly.params import player_marker_args
from src.pitchly.pitch import Pitch


class TrackingData:
    def __init__(self, data, metadata):
        self.metadata = metadata
        self.home_players = [x.player_id for x in self.metadata.teams[0].players]
        self.away_players = [x.player_id for x in self.metadata.teams[1].players]
        self.home_jerseys = [x.jersey_no for x in self.metadata.teams[0].players]
        self.away_jerseys = [x.jersey_no for x in self.metadata.teams[1].players]

        data = self.metric_coords(data)
        data = self.calc_player_velocities(data, filter_="moving average")
        data = self.flip_second_half_direction(data)
        self.data = data

    def metric_coords(self, data, field_dimen=FIELD_DIM):
        x_columns = [c for c in data.columns if c.endswith("_x")]
        y_columns = [c for c in data.columns if c.endswith("_y")]
        data[x_columns] = (data[x_columns] - 0.5) * field_dimen[0]
        data[y_columns] = -1 * (data[y_columns] - 0.5) * field_dimen[1]

        return data

    def calc_player_velocities(
        self, data, smoothing=True, filter_="moving average", window=7, polyorder=1, maxspeed=12
    ):
        # Get the player ids
        player_ids = self.home_players + self.away_players
        # Calculate the timestep from one frame to the next. Should always be 0.04 within the same half
        dt = data.timestamp.diff()

        # index of first frame in second half
        second_half_idx = data.period_id.idxmax(2)
        # estimate velocities for players in team
        for player in player_ids:  # cycle through players individually
            # difference player positions in timestep dt to get unsmoothed estimate of velicity
            vx = data[player + "_x"].diff() / dt
            vy = data[player + "_y"].diff() / dt

            if maxspeed > 0:
                # remove unsmoothed data points that exceed the maximum speed (these are most likely position errors)
                raw_speed = np.sqrt(vx ** 2 + vy ** 2)
                vx[raw_speed > maxspeed] = np.nan
                vy[raw_speed > maxspeed] = np.nan

            if smoothing:
                if filter_ == "Savitzky-Golay":
                    # calculate first half velocity
                    vx.iloc[:second_half_idx] = signal.savgol_filter(
                        vx.iloc[:second_half_idx], window_length=window, polyorder=polyorder
                    )
                    vy.iloc[:second_half_idx] = signal.savgol_filter(
                        vy.iloc[:second_half_idx], window_length=window, polyorder=polyorder
                    )
                    # calculate second half velocity
                    vx.iloc[second_half_idx:] = signal.savgol_filter(
                        vx.iloc[second_half_idx:], window_length=window, polyorder=polyorder
                    )
                    vy.iloc[second_half_idx:] = signal.savgol_filter(
                        vy.iloc[second_half_idx:], window_length=window, polyorder=polyorder
                    )
                elif filter_ == "moving average":
                    ma_window = np.ones(window) / window
                    # calculate first half velocity
                    vx.iloc[:second_half_idx] = np.convolve(vx.iloc[:second_half_idx], ma_window, mode="same")
                    vy.iloc[:second_half_idx] = np.convolve(vy.iloc[:second_half_idx], ma_window, mode="same")
                    # calculate second half velocity
                    vx.iloc[second_half_idx:] = np.convolve(vx.iloc[second_half_idx:], ma_window, mode="same")
                    vy.iloc[second_half_idx:] = np.convolve(vy.iloc[second_half_idx:], ma_window, mode="same")

            # put player speed in x,y direction, and total speed back in the data frame
            data[player + "_vx"] = vx
            data[player + "_vy"] = vy
            data[player + "_speed"] = np.sqrt(vx ** 2 + vy ** 2)

        return data

    def flip_second_half_direction(self, data):
        """
        Flip coordinates in second half so that each team always shoots in the same direction through the match.
        """
        second_half_idx = data.period_id.idxmax(2)
        columns = [c for c in data.columns if c[-1] in ["X", "y"]]
        data.loc[second_half_idx:, columns] *= -1
        return data

    def get_frameID_from_timestamp(self, timestamp):
        return self.data.query("timestamp==@timestamp").index[0]

    def get_timestamp(self, mins):
        if ":" in mins:
            seconds = int(mins.split(":")[0]) * 60 + int(mins.split(":")[1]) + 0.04
        else:
            seconds = int(mins) * 60
        return seconds

    def get_frame_data(self, frameID):

        frame_data = self.data.loc[frameID]

        return frame_data

    def position_traces(self, frame_data):
        player_ids = (self.home_players, self.away_players)
        jerseys = (self.home_jerseys, self.away_jerseys)

        position_traces = []
        for i, side in enumerate(["Home", "Away"]):
            players = player_ids[i]
            xlocs = [frame_data[f"{player_id}_x"] for player_id in players]
            ylocs = [frame_data[f"{player_id}_y"] for player_id in players]

            traces = go.Scatter(x=xlocs, y=ylocs, text=jerseys[i], **player_marker_args[side], name=side)
            position_traces.append(traces)

        return position_traces

    def velocity_traces(self, frame_data):
        player_ids = (self.home_players, self.away_players)
        velocity_quivers = []
        for i, side in enumerate(["Home", "Away"]):
            players = player_ids[i]
            xlocs = [frame_data[f"{player_id}_x"] for player_id in players]
            ylocs = [frame_data[f"{player_id}_y"] for player_id in players]
            xvels = [frame_data[f"{player_id}_vx"] for player_id in players]
            yvels = [frame_data[f"{player_id}_vy"] for player_id in players]

            trace = ff.create_quiver(
                x=xlocs,
                y=ylocs,
                u=xvels,
                v=yvels,
                scale=0.5,
                line_color=player_marker_args[side]["marker_color"],
                name=side + "_vel",
            )
            velocity_quivers.append(trace.data[0])

        return velocity_quivers

    def ball_trace(self, frame_data):
        ball_trace = go.Scatter(
            x=[frame_data["ball_x"]],
            y=[frame_data["ball_y"]],
            marker_size=10,
            marker_opacity=0.8,
            marker_color="black",
            marker_line_width=2,
            marker_line_color="green",
            name="ball",
        )

        return ball_trace

    def get_traces(self, frameID=None, velocities=True, ball=True):
        """Combines various traces for required plot and returns it

        Args:
            velocities (bool, optional): If True, velocity quivers will be added. Defaults to True.
            ball (bool, optional): If True, ball trace is added. Defaults to True.
        """
        frame_data = self.get_frame_data(frameID)

        traces = []
        if velocities:
            traces.extend(self.velocity_traces(frame_data))

        traces.extend(self.position_traces(frame_data))

        if ball:
            traces.append(self.ball_trace(frame_data))

        return traces

    def get_frames(self, frame_range, velocities=True, ball=True):

        frames = []
        for frameID in tqdm(frame_range):
            frame = {"data": [], "name": str(frameID)}
            frame["data"].extend(self.get_traces(frameID, velocities, ball))
            frames.append(frame)

        return frames

    def plot_frame(self, frameID=None, time=None, plot_ball=True, show_velocities=False):

        if time:
            seconds = self.get_timestamp(time)
            frameID = self.get_frameID_from_timestamp(seconds)
        else:
            seconds = self.data.loc[frameID, "timestamp"]
            time = f"{seconds//60:0.0f}'{seconds%60:0.0f}\""

        title = f"Time: [{time}] | FrameID: {frameID}"
        data = self.get_traces(frameID=frameID, velocities=show_velocities, ball=plot_ball)
        pitch = Pitch()
        return pitch.plot_freeze_frame(data, title)

    def plot_sequence(self, f0=None, f1=None, t0=None, t1=None, show_velocities=True, player_num=None):

        if t1:
            t0 = self.get_timestamp(t0)
            t1 = self.get_timestamp(t1)

            f0 = self.get_frameID_from_timestamp(t0)
            f1 = self.get_frameID_from_timestamp(t1)
        else:
            seconds = self.data.loc[f0, "timestamp"]
            t0 = f"{seconds//60:0.0f}'{seconds%60:0.0f}\""

        frame_range = range(f0, f1)

        title = f"Time: [{t0}] | FrameID: {f0} to {f1}"
        data = self.get_traces(frameID=f0, velocities=show_velocities)
        frames = self.get_frames(frame_range, velocities=show_velocities)
        pitch = Pitch()
        return pitch.plot_frames_sequence(data, frames, frame_range, title)


class EventData:
    def __init__(self, events: list):
        self.events = self.add_tags(events)
        for event in self.events:
            tags = event.raw_event["tags"]
            if "SAVED" in tags or "BLOCKED" in tags:
                event.raw_event["marker_color"] = "white"
                event.raw_event["marker_line_color"] = "white"
                event.raw_event["marker_line_width"] = 2

                if "HEAD" in tags:
                    event.raw_event["marker_symbol"] = "circle"
                else:
                    event.raw_event["marker_symbol"] = "triangle-up"

            elif getattr(event.result, "value", "") == "GOAL":
                event.raw_event["marker_color"] = "darkgreen"
                event.raw_event["marker_line_color"] = "white"
                event.raw_event["marker_line_width"] = 2

                if "HEAD" in tags:
                    event.raw_event["marker_symbol"] = "circle"
                else:
                    event.raw_event["marker_symbol"] = "triangle-up"

            else:
                event.raw_event["marker_color"] = "yellow"
                event.raw_event["marker_line_color"] = "white"
                event.raw_event["marker_line_width"] = 2

                if "HEAD" in tags:
                    event.raw_event["marker_symbol"] = "circle-open"
                else:
                    event.raw_event["marker_symbol"] = "triangle-up-open"

        # self.events = events

    def metric_coords(self, event_list):
        for event in event_list:
            # flip sign
            sign = 1 if event.period.id == 1 else -1
            # print(event.raw_event)
            # transform to metric dim (106,68)
            event.raw_event["start"]["X"] = sign * ((event.raw_event["start"]["x"] - 0.5) * 106.0)
            event.raw_event["start"]["Y"] = sign * (-1 * (event.raw_event["start"]["y"] - 0.5) * 68.0)
            if event.raw_event["end"]["x"] is not None:
                event.raw_event["end"]["X"] = sign * ((event.raw_event["end"]["x"] - 0.5) * 106.0)
                event.raw_event["end"]["Y"] = sign * (-1 * (event.raw_event["end"]["y"] - 0.5) * 68.0)
        return event_list

    def add_tags(self, event_list):
        for event in event_list:
            subtypes = event.raw_event.get("subtypes", "")
            tags = []
            if subtypes:
                if isinstance(subtypes, list):
                    tags.extend([x["name"] for x in subtypes])
                else:
                    tags.append(subtypes["name"])
            event.raw_event["tags"] = tags
        return event_list

    def get_shots(self):
        shots = [event for event in self.events if event.event_name == "shot"]
        shots = self.metric_coords(shots)
        return shots

    def get_goals(self):
        goals = [event for event in self.events if getattr(event.result, "value", "") == "GOAL"]
        goals = self.metric_coords(goals)
        return goals

    def get_corners(self):
        corners = []
        for i, event in enumerate(self.events):
            if "CORNER KICK" in event.raw_event.get("tags"):
                corners.append(self.events[i + 1])
        corners = self.metric_coords(corners)
        return corners

    def get_freekicks(self):
        freekicks = []
        for i, event in enumerate(self.events):
            if "FREE KICK" in event.raw_event.get("tags"):
                freekicks.append(self.events[i + 1])
        freekicks = self.metric_coords(freekicks)
        return freekicks

    def get_challenges(self):
        challenges = [event for event in self.events if event.event_name == "CHALLENGE"]
        challenges = self.metric_coords(challenges)
        return challenges

    def get_recoveries(self):
        recoveries = [
            event
            for event in self.events
            if event.event_name == "recovery"
            and ("INTERCEPTION" in event.raw_event["tags"] or "THEFT" in event.raw_event["tags"])
        ]
        recoveries = self.metric_coords(recoveries)
        return recoveries

    def get_turnovers(self):
        turnovers = [
            event
            for event in self.events
            if event.event_name == "pass"
            and ("INTERCEPTION" in event.raw_event["tags"] or "THEFT" in event.raw_event["tags"])
        ]
        turnovers = self.metric_coords(turnovers)
        return turnovers

    def get_shot_assists(self):
        shot_assists = []  # passes that lead to SHOT/CARRY+SHOT
        for i, event in enumerate(self.events):
            if event.event_name == "shot":
                if self.events[i - 1].event_name == "pass":
                    # passes that led to shots
                    shot_assists.append(self.events[i - 1])
                elif self.events[i - 1].event_name == "carry" and self.events[i - 2].event_name == "pass":
                    # passes that led to carry+shot
                    shot_assists.append(self.events[i - 2])

        shot_assists = self.metric_coords(shot_assists)
        return shot_assists

    def get_assists(self):
        assists = []  # passes that lead to GOAL/CARRY+GOAL
        for i, event in enumerate(self.events):
            if getattr(event.result, "value", "") == "GOAL":
                if self.events[i - 1].event_name == "pass":
                    # passes that led to goals
                    assists.append(self.events[i - 1])
                elif self.events[i - 1].event_name == "carry" and self.events[i - 2].event_name == "pass":
                    # passes that led to carry+goal
                    assists.append(self.events[i - 2])

        assists = self.metric_coords(assists)
        return assists

    def get_passes(self):
        # only complete passes for now
        passes = [event for event in self.events if event.event_name == "pass" and event.result.value == "COMPLETE"]
        passes = self.metric_coords(passes)
        return passes

    def get_buildup(self, index):
        # PASS/CARRY , ![GENERIC CHALLENGE/RECOVERY]
        buildup = []
        for i, event in enumerate(self.events):
            if event.raw_event["index"] == index:
                team = event.ball_owning_team
                while i > 0:
                    # print (i,self.events[i].raw_event['index'])
                    if self.events[i].result:
                        buildup.append(self.events[i])
                    i -= 1
                    if self.events[i].ball_owning_team and self.events[i].ball_owning_team != team:
                        break

        return buildup

    def event_traces(self, row):
        event_traces = []
        x, y, player_nums = [], [], []
        x.extend([row.raw_event["start"]["X"], row.raw_event["end"]["X"]])
        y.extend([row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]])
        if row.event_name == "pass":
            player_nums.extend([row.player.jersey_no, row.receiver_player.jersey_no])
        else:
            player_nums.extend([row.player.jersey_no, None])

        event_traces.append(
            go.Scatter(
                x=x,
                y=y,
                text=player_nums,
                name=row.event_type.name,
                mode="lines+markers+text",
                marker_size=[20, 0],
                # marker_symbol="circle-x",
                marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                marker_line_color="white",
                marker_line_width=[2, 0],
                line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                line_width=1,
                line_dash="dash" if row.event_name == "carry" else None,
                textfont=dict(size=11, color="white"),
                showlegend=False,
            )
        )

        return event_traces

    def get_traces(self, index):
        event_chain = self.get_buildup(index)
        event_chain = self.metric_coords(event_chain)
        traces = []
        for row in event_chain:
            traces.extend(self.event_traces(row))

        return traces

    def title(self, index):
        for row in self.events:
            text = "Event"
            raw = row.raw_event
            if raw["index"] == index:
                time = f"{raw['start']['time']//60:0.0f}'{raw['start']['time']%60:0.0f}\""
                text = f"{raw['type']['name']} > {raw['tags']} > {row.team} | {row.player.name} @ {time}"
                break

        return text

    def plot(self, index=None, type=None, team=None, player=None, trace=False):

        if index:
            pitch = Pitch()
            return pitch.plot_event(data=self.get_traces(index), title=self.title(index))

        if type == "shots":
            data = self.get_shots()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], row.raw_event["end"]["X"]],
                        y=[row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]],
                        text=f"{row.result.value}<br>{row.player}({row.team})",  # [None, None],
                        name=row.result.value,
                        mode="lines+markers" if trace else "markers",
                        marker_size=[15, 0],
                        marker_symbol=row.raw_event["marker_symbol"],
                        marker_color=row.raw_event["marker_color"],
                        marker_line_color=row.raw_event["marker_line_color"],
                        marker_line_width=[row.raw_event["marker_line_width"], 0],
                        line_color=row.raw_event["marker_line_color"],
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "goals":
            data = self.get_goals()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], row.raw_event["end"]["X"]],
                        y=[row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]],
                        text=f"{row.result.value}<br>{row.player}({row.team})",  # [None, None],
                        name=row.result.value,
                        mode="lines+markers" if trace else "markers",
                        marker_size=[15, 0],
                        marker_symbol=row.raw_event["marker_symbol"],
                        marker_color=row.raw_event["marker_color"],
                        marker_line_color=row.raw_event["marker_line_color"],
                        marker_line_width=[row.raw_event["marker_line_width"], 0],
                        line_color=row.raw_event["marker_line_color"],
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "corners":
            data = self.get_corners()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], row.raw_event["end"]["X"]],
                        y=[row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]],
                        text=f"{row.result.value}<br>{row.player}({row.team})",  # [None, None],
                        name=row.result.value,
                        mode="lines+markers" if trace else "markers",
                        marker_size=[0, 10],
                        marker_symbol="square",
                        marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[0, 2],
                        line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "freekicks":
            data = self.get_freekicks()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], row.raw_event["end"]["X"]],
                        y=[row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]],
                        text=f"{row.result.value}<br>{row.player}({row.team})",  # [None, None],
                        name=row.result.value,
                        mode="lines+markers" if trace else "markers",
                        marker_size=[15, 0],
                        marker_symbol="circle-x",
                        marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[2, 0],
                        line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "challenges":
            data = self.get_challenges()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], None],
                        y=[row.raw_event["start"]["Y"], None],
                        text=f"{row.raw_event['tags']}<br>{row.player}({row.team})",  # [None, None],
                        name=row.raw_event["tags"][0],
                        mode="markers",
                        marker_size=[18, 0],
                        marker_symbol="hexagon",
                        marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[1, 0],
                        line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "recoveries":
            data = self.get_recoveries()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], None],
                        y=[row.raw_event["start"]["Y"], None],
                        text=f"{row.raw_event['tags']}<br>{row.player}({row.team})",  # [None, None],
                        name=row.raw_event["tags"][0],
                        mode="markers",
                        marker_size=[18, 0],
                        marker_symbol="hexagon",
                        marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[1, 0],
                        line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "turnovers":
            data = self.get_turnovers()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], None],
                        y=[row.raw_event["start"]["Y"], None],
                        text=f"{row.raw_event['tags']}<br>{row.player}({row.team})",  # [None, None],
                        name=row.raw_event["tags"][0],
                        mode="markers",
                        marker_size=[18, 0],
                        marker_symbol="hexagon",
                        marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[1, 0],
                        line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "shot_assists":
            data = self.get_shot_assists()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], row.raw_event["end"]["X"]],
                        y=[row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]],
                        text=[row.player.jersey_no, None],
                        name=row.result.value,
                        mode="lines+markers+text" if trace else "markers+text",
                        marker_size=[20, 0],
                        marker_symbol="pentagon",
                        marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[2, 0],
                        line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "assists":
            data = self.get_assists()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], row.raw_event["end"]["X"]],
                        y=[row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]],
                        text=[row.player.jersey_no, None],
                        name=row.result.value,
                        mode="lines+markers+text" if trace else "markers+text",
                        marker_size=[20, 0],
                        marker_symbol="pentagon",
                        marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[2, 0],
                        line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "passes":
            data = self.get_passes()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], row.raw_event["end"]["X"]],
                        y=[row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]],
                        text=[row.player.jersey_no, row.receiver_player.jersey_no],
                        name=row.result.value,
                        mode="lines",
                        # marker_size=[0, 0],
                        # marker_symbol=row["marker_symbol"],
                        # marker_color=row["marker_color"],
                        # marker_line_color=row["marker_line_color"],
                        # marker_line_width=[row["marker_line_width"], 0],
                        line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        line_width=1,
                        # textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "passers":
            data = self.get_passes()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], row.raw_event["end"]["X"]],
                        y=[row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]],
                        text=[None, None],
                        name=row.result.value,
                        mode="markers+text",
                        marker_size=[10, 0],
                        marker_symbol="diamond",
                        marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[1, 0],
                        # line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        # line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "receivers":
            data = self.get_passes()
            traces = []
            for row in data:
                traces.append(
                    go.Scatter(
                        x=[row.raw_event["start"]["X"], row.raw_event["end"]["X"]],
                        y=[row.raw_event["start"]["Y"], row.raw_event["end"]["Y"]],
                        text=[None, None],
                        name=row.result.value,
                        mode="markers+text",
                        marker_size=[0, 10],
                        marker_symbol="diamond",
                        marker_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[0, 1],
                        # line_color="#AD0B05" if row.team.ground.name == "HOME" else "#0570B0",
                        # line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        pitch = Pitch()
        return pitch.plot_event(data=traces, title=type)



import glob

import numpy as np
import pandas as pd
# import plotly.graph_objects as go
import scipy.signal as signal

# from src.pitchly.params import FIELD_COLOR
from src.pitchly.params import FIELD_DIM

# from collections import namedtuple
# from src.pitchly.params import FIELD_HEIGHT
# from src.pitchly.params import FIELD_MARKINGS_COLOR
# from src.pitchly.params import FIELD_WIDTH
# from src.pitchly.params import player_marker_args

match_dir = "/media/opunsoars/My Book/playground/friends_of_tracking/friends_of_tracking/\
    datahub/metrica_sports/sample-data/data/Sample_Game_1"


def modify_cols(tracking_df):
    cols = list(tracking_df.iloc[2, :3])
    # cols = cols[:3]
    for i in range(3, tracking_df.shape[1] - 2, 2):
        cols.append(f"{tracking_df.iloc[0,i]}_{tracking_df.iloc[1,i]}_X")
        cols.append(f"{tracking_df.iloc[0,i]}_{tracking_df.iloc[1,i]}_Y")
    cols.append("ball_X")
    cols.append("ball_Y")

    tracking_df.columns = cols
    tracking_df = tracking_df.loc[3:, :]
    for col in cols[2:]:
        tracking_df[col] = tracking_df[col].astype(float)
    tracking_df["Frame"] = tracking_df.Frame.astype(int)
    tracking_df["Period"] = tracking_df.Period.astype(int)
    tracking_df["mins"] = tracking_df["Time [s]"].apply(lambda x: f"{x//60:0.0f}:{x%60:0.2f}")
    tracking_df.set_index("Frame", inplace=True)
    return tracking_df


def convert_to_metric_coords(data, field_dimen=FIELD_DIM):
    """
    Convert positions from Metrica units to meters (with origin at centre circle)
    """
    x_columns = [c for c in data.columns if c.endswith("X")]
    y_columns = [c for c in data.columns if c.endswith("Y")]
    data[x_columns] = (data[x_columns] - 0.5) * field_dimen[0]
    data[y_columns] = -1 * (data[y_columns] - 0.5) * field_dimen[1]
    """
    ------------ ***NOTE*** ------------
    Metrica actually define the origin at the *top*-left of the field, not the bottom-left, as discussed in the YouTube video.
    I've changed the line above to reflect this. It was originally:
    data[y_columns] = ( data[y_columns]-0.5 ) * field_dimen[1]
    ------------ ********** ------------
    """
    return data


def calc_player_velocities(team, smoothing=True, filter_="moving average", window=7, polyorder=1, maxspeed=12):
    """calc_player_velocities( tracking_data )

    Calculate player velocities in x & y direciton, and total player speed at each timestamp of the tracking data

    Parameters
    -----------
        team: the tracking DataFrame for home or away team
        smoothing: boolean variable that determines whether velocity measures are smoothed. Default is True.
        filter: type of filter to use when smoothing the velocities. Default is Savitzky-Golay,\
             which fits a polynomial of order 'polyorder' to the data within each window
        window: smoothing window size in # of frames
        polyorder: order of the polynomial for the Savitzky-Golay filter. \
            Default is 1 - a linear fit to the velcoity, so gradient is the acceleration
        maxspeed: the maximum speed that a player can realisitically achieve (in meters/second). \
            d measures that exceed maxspeed are tagged as outliers and set to NaN.

    Returrns
    -----------
       team : the tracking DataFrame with columns for speed in the x & y direction and total speed added

    """
    # remove any velocity data already in the dataframe
    team = remove_player_velocities(team)
    # print (team.isna().sum())

    # Get the player ids
    player_ids = np.unique([c[:-2] for c in team.columns if c[:4] in ["Home", "Away"]])

    # Calculate the timestep from one frame to the next. Should always be 0.04 within the same half
    dt = team["Time [s]"].diff()

    # index of first frame in second half
    second_half_idx = team.Period.idxmax(2)

    # estimate velocities for players in team
    for player in player_ids:  # cycle through players individually
        # difference player positions in timestep dt to get unsmoothed estimate of velicity
        vx = team[player + "_X"].diff() / dt
        vy = team[player + "_Y"].diff() / dt

        if maxspeed > 0:
            # remove unsmoothed data points that exceed the maximum speed (these are most likely position errors)
            raw_speed = np.sqrt(vx ** 2 + vy ** 2)
            vx[raw_speed > maxspeed] = np.nan
            vy[raw_speed > maxspeed] = np.nan

        if smoothing:
            if filter_ == "Savitzky-Golay":
                # calculate first half velocity
                vx.iloc[:second_half_idx] = signal.savgol_filter(
                    vx.iloc[:second_half_idx], window_length=window, polyorder=polyorder
                )
                vy.iloc[:second_half_idx] = signal.savgol_filter(
                    vy.iloc[:second_half_idx], window_length=window, polyorder=polyorder
                )
                # calculate second half velocity
                vx.iloc[second_half_idx:] = signal.savgol_filter(
                    vx.iloc[second_half_idx:], window_length=window, polyorder=polyorder
                )
                vy.iloc[second_half_idx:] = signal.savgol_filter(
                    vy.iloc[second_half_idx:], window_length=window, polyorder=polyorder
                )
            elif filter_ == "moving average":
                ma_window = np.ones(window) / window
                # calculate first half velocity
                vx.iloc[:second_half_idx] = np.convolve(vx.iloc[:second_half_idx], ma_window, mode="same")
                vy.iloc[:second_half_idx] = np.convolve(vy.iloc[:second_half_idx], ma_window, mode="same")
                # calculate second half velocity
                vx.iloc[second_half_idx:] = np.convolve(vx.iloc[second_half_idx:], ma_window, mode="same")
                vy.iloc[second_half_idx:] = np.convolve(vy.iloc[second_half_idx:], ma_window, mode="same")

        # put player speed in x,y direction, and total speed back in the data frame
        team[player + "_vx"] = vx
        team[player + "_vy"] = vy
        team[player + "_speed"] = np.sqrt(vx ** 2 + vy ** 2)

    return team


def remove_player_velocities(team):
    # remove player velocoties and acceleeration measures that are already in the 'team' dataframe
    columns = [
        c for c in team.columns if c.split("_")[-1] in ["vx", "vy", "ax", "ay", "speed", "acceleration"]
    ]  # Get the player ids
    team = team.drop(columns=columns)
    return team


def flip_second_half_direction(team):
    """
    Flip coordinates in second half so that each team always shoots in the same direction through the match.
    """
    second_half_idx = team.Period.idxmax(2)
    columns = [c for c in team.columns if c[-1].lower() in ["x", "y"]]
    team.loc[second_half_idx:, columns] *= -1
    return team


def load_data(match_dir):

    home_track = (
        pd.read_csv(glob.glob(f"{match_dir}/*Home*.csv")[0], header=None)
        .pipe(modify_cols)
        .pipe(convert_to_metric_coords)
        .pipe(calc_player_velocities)
        .pipe(flip_second_half_direction)
    )

    away_track = (
        pd.read_csv(glob.glob(f"{match_dir}/*Away*.csv")[0], header=None)
        .pipe(modify_cols)
        .pipe(convert_to_metric_coords)
        .pipe(calc_player_velocities)
        .pipe(flip_second_half_direction)
    )

    events = (
        pd.read_csv(glob.glob(f"{match_dir}/*Events*.csv")[0])
        .pipe(convert_to_metric_coords)
        .pipe(flip_second_half_direction)
    )

    return home_track, away_track, events


# tracking_home, tracking_away, events = load_data(match_dir)



import re

import numpy as np
import plotly.figure_factory as ff
import plotly.graph_objects as go
from tqdm.auto import tqdm

# from src.pitchly.params import FIELD_COLOR
# from src.pitchly.params import FIELD_DIM
# from src.pitchly.params import FIELD_HEIGHT
# from src.pitchly.params import FIELD_MARKINGS_COLOR
# from src.pitchly.params import FIELD_WIDTH
from src.pitchly.params import event_player_marker_args
from src.pitchly.params import player_marker_args
from src.pitchly.pitch import Pitch


class TrackingData:
    def __init__(self, tracking_home, tracking_away):
        self.tracking_home = tracking_home
        self.tracking_away = tracking_away

    def get_frame_data(self, frameID):

        hometeam = self.tracking_home.loc[frameID]
        awayteam = self.tracking_away.loc[frameID]
        frame_data = (hometeam, awayteam)

        return frame_data

    def get_mins(self, time):
        if ":" not in time:
            return f"{time}:0.04"
        else:
            return f"{time}.04"

    def get_frameID_from_mins(self, mins):
        return self.tracking_home.query("mins==@mins").index[0]

    def position_traces(self, frame_data):

        position_traces = []
        for i, side in enumerate(["Home", "Away"]):
            team_data = frame_data[i]
            player_nums = list(
                set(item for subitem in team_data.keys() for item in subitem.split("_") if item.isdigit())
            )
            xlocs = [team_data["{}_{}_{}".format(side, num, "X")] for num in player_nums]
            ylocs = [team_data["{}_{}_{}".format(side, num, "Y")] for num in player_nums]
            traces = go.Scatter(x=xlocs, y=ylocs, text=player_nums, **player_marker_args[side], name=side)
            position_traces.append(traces)

        return position_traces

    def velocity_traces(self, frame_data):

        velocity_quivers = []
        for i, side in enumerate(["Home", "Away"]):
            team_data = frame_data[i]
            player_nums = list(
                set(item for subitem in team_data.keys() for item in subitem.split("_") if item.isdigit())
            )
            xlocs = [team_data["{}_{}_{}".format(side, num, "X")] for num in player_nums]
            ylocs = [team_data["{}_{}_{}".format(side, num, "Y")] for num in player_nums]
            xvels = [team_data["{}_{}_{}".format(side, num, "vx")] for num in player_nums]
            yvels = [team_data["{}_{}_{}".format(side, num, "vy")] for num in player_nums]
            trace = ff.create_quiver(
                x=xlocs,
                y=ylocs,
                u=xvels,
                v=yvels,
                scale=0.5,
                line_color=player_marker_args[side]["marker_color"],
                name=side + "_vel",
            )
            velocity_quivers.append(trace.data[0])

        return velocity_quivers

    def ball_trace(self, frame_data):

        team_data = frame_data[0]
        ball_trace = go.Scatter(
            x=[team_data["ball_X"]],
            y=[team_data["ball_Y"]],
            marker_size=10,
            marker_opacity=0.8,
            marker_color="black",
            marker_line_width=2,
            marker_line_color="green",
            name="ball",
        )

        return ball_trace

    def get_traces(self, frameID=None, velocities=True, ball=True):
        """Combines various traces for required plot and returns it

        Args:
            velocities (bool, optional): If True, velocity quivers will be added. Defaults to True.
            ball (bool, optional): If True, ball trace is added. Defaults to True.
        """
        frame_data = self.get_frame_data(frameID)

        traces = []
        if velocities:
            traces.extend(self.velocity_traces(frame_data))

        traces.extend(self.position_traces(frame_data))

        if ball:
            traces.append(self.ball_trace(frame_data))

        return traces

    def get_frames(self, frame_range, velocities=True, ball=True):

        frames = []
        for frameID in tqdm(frame_range):
            frame = {"data": [], "name": str(frameID)}
            frame["data"].extend(self.get_traces(frameID, velocities, ball))
            frames.append(frame)

        return frames

    def plot_frame(self, frameID=None, time=None, plot_ball=True, show_velocities=False):

        if time:
            time = self.get_mins(time)
            frameID = self.get_frameID_from_mins(time)
        else:
            time = self.tracking_home.loc[frameID, "mins"]

        title = f"Time: [{time}] | FrameID: {frameID}"
        data = self.get_traces(frameID=frameID, velocities=show_velocities, ball=plot_ball)
        pitch = Pitch()
        return pitch.plot_freeze_frame(data, title)

    def plot_sequence(self, f0=None, f1=None, t0=None, t1=None, show_velocities=True, player_num=None):

        if t1:
            t0 = self.get_mins(t0)
            t1 = self.get_mins(t1)

            f0 = self.get_frameID_from_mins(t0)
            f1 = self.get_frameID_from_mins(t1)
        else:
            t0 = self.tracking_home.loc[f0, "mins"]

        frame_range = range(f0, f1)

        title = f"Time: [{t0}] | FrameID: {f0} to {f1}"
        data = self.get_traces(frameID=f0, velocities=show_velocities)
        frames = self.get_frames(frame_range, velocities=show_velocities)
        pitch = Pitch()
        return pitch.plot_frames_sequence(data, frames, frame_range, title)


class EventData:
    def __init__(self, events):
        self.events = events
        self.is_shot = self.events.Type == "SHOT"
        self.is_goal = self.events.Subtype.str.endswith("GOAL", na=False)
        self.is_saved = self.events.Subtype.str.endswith("SAVED", na=False)
        self.is_pass = self.events.Type == "PASS"
        self.is_fk = self.events.Subtype == "FREE KICK"
        self.is_corner = self.events.Subtype == "CORNER KICK"
        self.is_balllost = self.events.Type == "BALL LOST"
        self.is_challenge = self.events.Type == "CHALLENGE"

        for i in self.events.index:
            if not isinstance(self.events.loc[i, "Subtype"], str):
                continue
            # if self.events["Type"].iloc[i] == "SHOT":
            if self.events["Subtype"].iloc[i].endswith("SAVED") or self.events["Subtype"].iloc[i].endswith("BLOCKED"):
                self.events.loc[i, "marker_color"] = "white"
                self.events.loc[i, "marker_line_color"] = "white"
                self.events.loc[i, "marker_line_width"] = 2

                if self.events["Subtype"].iloc[i].startswith("HEAD"):
                    self.events.loc[i, "marker_symbol"] = "circle"
                else:
                    self.events.loc[i, "marker_symbol"] = "triangle-up"

            elif self.events.loc[i, "Subtype"].endswith("GOAL"):
                self.events.loc[i, "marker_color"] = "darkgreen"
                self.events.loc[i, "marker_line_color"] = "white"
                self.events.loc[i, "marker_line_width"] = 2

                if self.events["Subtype"].iloc[i].startswith("HEAD"):
                    self.events.loc[i, "marker_symbol"] = "circle"
                else:
                    self.events.loc[i, "marker_symbol"] = "triangle-up"

            else:
                self.events.loc[i, "marker_color"] = "yellow"
                self.events.loc[i, "marker_line_color"] = "white"
                self.events.loc[i, "marker_line_width"] = 2

                if self.events["Subtype"].iloc[i].startswith("HEAD"):
                    self.events.loc[i, "marker_symbol"] = "circle-open"
                else:
                    self.events.loc[i, "marker_symbol"] = "triangle-up-open"

    def get_passes(self):
        return self.events[self.is_pass]

    def get_shots(self):
        return self.events[self.is_shot]

    def get_goals(self):
        return self.events[self.is_goal]

    def get_corners(self):
        corner_index = self.events[self.is_corner].index

        return self.events.iloc[corner_index + 1]

    def get_freekicks(self):
        fk_index = self.events[self.is_fk].index

        return self.events.iloc[fk_index + 1]

    def get_challenges(self):
        return self.events[self.is_challenge]

    def get_turnovers(self):
        return self.events[self.is_balllost]

    def get_shot_assists(self):
        shots = self.get_shots()
        assist_index = [
            i - 1
            for i in shots.index
            if (self.events.Type.iloc[i - 1] == "PASS")
            or (self.events.Subtype.iloc[i - 1] == "CORNER KICK")
            or (self.events.Subtype.iloc[i - 1] == "FREE KICK")
        ]
        return self.events.iloc[assist_index]

    def get_assists(self):
        goals = self.get_goals()
        assist_index = [
            i - 1
            for i in goals.index
            if (self.events.Type.iloc[i - 1] == "PASS")
            or (self.events.Subtype.iloc[i - 1] == "CORNER KICK")
            or (self.events.Subtype.iloc[i - 1] == "FREE KICK")
        ]
        return self.events.iloc[assist_index]

    def title(self, index):
        row = self.events.iloc[index]
        time = f"{row['Start Time [s]']//60:0.0f}'{row['Start Time [s]']%60:0.0f}\""
        text = f"{row['Type']} > {row['Subtype']} > {row['Team']} {row['From']} @ {time}"
        return text

    def get_buildup(self, index):
        for i, row in self.events.iloc[index - 1 :: -1, :2].iterrows():
            if row["Type"] != "PASS":
                break
        buildup = self.events.iloc[i + 1 : index + 1, :]
        buildup["From_num"] = buildup.From.apply(lambda x: re.findall("\d+", str(x))).apply(
            lambda x: x[0] if len(x) > 0 else None
        )
        buildup["To_num"] = buildup.To.apply(lambda x: re.findall(r"\d+", str(x))).apply(
            lambda x: x[0] if len(x) > 0 else None
        )
        buildup["To_prev"] = buildup.To_num.shift(1)
        buildup["Start X Carry"] = buildup["End X"]
        buildup["Start Y Carry"] = buildup["End Y"]
        buildup["End X Carry"] = buildup["Start X"].shift(-1)
        buildup["End Y Carry"] = buildup["Start Y"].shift(-1)
        buildup["carry"] = (buildup["Start X Carry"] != buildup["End X Carry"]) | (
            buildup["Start Y Carry"] != buildup["End Y Carry"]
        )
        buildup.loc[buildup["carry"] == False, "Start X Carry"] = np.nan
        buildup.loc[buildup["carry"] == False, "Start Y Carry"] = np.nan
        buildup.loc[buildup["carry"] == False, "End X Carry"] = np.nan
        buildup.loc[buildup["carry"] == False, "End Y Carry"] = np.nan

        buildup["Start Frame Carry"] = buildup["End Frame"]
        buildup["End Frame Carry"] = buildup["Start Frame"].shift(-1)

        return buildup

    def event_traces(self, row):
        event_traces = []
        x, y, player_nums = [], [], []
        x.extend([row["Start X"], row["End X"]])
        y.extend([row["Start Y"], row["End Y"]])
        side = row["Team"]
        player_nums.extend([row["From_num"], row["To_num"]])

        event_traces.append(go.Scatter(x=x, y=y, text=player_nums, name=row["Type"], **event_player_marker_args[side]))

        if (row["Type"] == "PASS") & (row["carry"] == True):
            event_traces.append(
                go.Scatter(
                    x=[row["Start X Carry"], row["End X Carry"]],
                    y=[row["Start Y Carry"], row["End Y Carry"]],
                    text=[row["To_num"]],
                    name="DRIBBLE",
                    line_dash="dash",
                    **event_player_marker_args[side],
                )
            )

        return event_traces

    def get_traces(self, index):
        event_chain = self.get_buildup(index)
        traces = []
        for i, row in event_chain.iterrows():
            traces.extend(self.event_traces(row))

        return traces

    def plot(self, index=None, type=None, team=None, player=None, trace=False):

        if index:
            pitch = Pitch()
            return pitch.plot_event(data=self.get_traces(index), title=self.title(index))

        if type == "shots":
            data = self.get_shots()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[None, None],
                        name=row["Type"],
                        mode="lines+markers+text" if trace else "markers+text",
                        marker_size=[15, 0],
                        marker_symbol=row["marker_symbol"],
                        marker_color=row["marker_color"],
                        marker_line_color=row["marker_line_color"],
                        marker_line_width=[row["marker_line_width"], 0],
                        line_color=row["marker_line_color"],
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "goals":
            data = self.get_goals()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[None, None],
                        name=row["Type"],
                        mode="lines+markers+text" if trace else "markers+text",
                        marker_size=[15, 0],
                        marker_symbol=row["marker_symbol"],
                        marker_color=row["marker_color"],
                        marker_line_color=row["marker_line_color"],
                        marker_line_width=[row["marker_line_width"], 0],
                        line_color=row["marker_line_color"],
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "freekicks":
            data = self.get_freekicks()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[None, None],
                        name=row["Type"],
                        mode="lines+markers+text" if trace else "markers+text",
                        marker_size=[15, 0],
                        marker_symbol="circle-x",
                        marker_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[2, 0],
                        line_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "assists":
            data = self.get_assists()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[row["From"].split("Player")[1], None],
                        name=row["Type"],
                        mode="lines+markers+text" if trace else "markers+text",
                        marker_size=[18, 0],
                        marker_symbol="pentagon",
                        marker_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[2, 0],
                        line_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "shot_assists":
            data = self.get_shot_assists()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[row["From"].split("Player")[1], None],
                        name=row["Type"],
                        mode="lines+markers+text" if trace else "markers+text",
                        marker_size=[18, 0],
                        marker_symbol="pentagon",
                        marker_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[2, 0],
                        line_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "corners":
            data = self.get_corners()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[None, None],
                        name=row["Type"],
                        mode="lines+markers+text" if trace else "markers+text",
                        marker_size=[0, 10],
                        marker_symbol="square",
                        marker_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[0, 2],
                        line_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "challenges":
            data = self.get_challenges()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[None, None],
                        name=row["Subtype"],
                        mode="markers+text",
                        marker_size=[18, 0],
                        marker_symbol="hexagon",
                        marker_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[1, 0],
                        line_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "turnovers":
            data = self.get_turnovers()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[None, None],
                        name=row["Subtype"],
                        mode="markers+text",
                        marker_size=[18, 0],
                        marker_symbol="hexagon",
                        marker_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[1, 0],
                        line_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "passes":
            data = self.get_passes()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[None, None],
                        name=row["Type"],
                        mode="lines",
                        # marker_size=[0, 0],
                        # marker_symbol=row["marker_symbol"],
                        # marker_color=row["marker_color"],
                        # marker_line_color=row["marker_line_color"],
                        # marker_line_width=[row["marker_line_width"], 0],
                        line_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        line_width=1,
                        # textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "passers":
            data = self.get_passes()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[None, None],
                        name=row["Type"],
                        mode="markers+text",
                        marker_size=[12, 0],
                        marker_symbol="diamond",
                        marker_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[1, 0],
                        # line_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        # line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        elif type == "receivers":
            data = self.get_passes()
            traces = []
            for i, row in data.iterrows():
                traces.append(
                    go.Scatter(
                        x=[row["Start X"], row["End X"]],
                        y=[row["Start Y"], row["End Y"]],
                        text=[None, None],
                        name=row["Type"],
                        mode="markers+text",
                        marker_size=[0, 12],
                        marker_symbol="diamond",
                        marker_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        marker_line_color="white",
                        marker_line_width=[0, 1],
                        # line_color="#AD0B05" if row["Team"] == "Home" else "#0570B0",
                        # line_width=1,
                        textfont=dict(size=11, color="white"),
                        showlegend=False,
                    )
                )

        pitch = Pitch()
        return pitch.plot_event(data=traces, title=type)
